#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# @author: melomcr


from networkx.algorithms.shortest_paths.dense import floyd_warshall as nxFW
from networkx.algorithms.shortest_paths.dense import floyd_warshall_predecessor_and_distance as nxFWPD
from networkx                       import edge_betweenness_centrality as nxbetweenness

from collections    import OrderedDict

import copy
import numpy as np

def calcOptPathPar(nxGraphs, inQueue, outQueue):
    '''Wrapper to calculate Floyd Warshall optimal path in parallel.
    
    For the FW optimal path determination, we use the node "distance" as weights, 
    that is, the log-transformation of the correlations, NOT the correlation itself.
    
    For every window, the function stores in the output queue the optimal paths. It turns the dictionary of distances returned by NetworkX into a NumPy 2D array per window of trajectory, which allows significant speed up in data analysis.
    
    Args:
        nxGraphs (obj) : NetworkX object.
        inQueue (obj) : Multiprocessing queue object for acquiring jobs.
        outQueue (obj) : Multiprocessing queue object for placing results.
    
    '''
    
    while not inQueue.empty():
        
        win = inQueue.get()
        
        ### IMPORTANT!
        # For the FW optimal path determination, we use the "distance" as weight, 
        #  that is, the log-transformation of the correlations. NOT the correlation itself.
        pathsPred, pathsDist = nxFWPD(nxGraphs[win], weight='dist')

        # Turns dictionary of distances into NumPy 2D array per window (for speed up)
        # Notice the nested list comprehensions.
        dists = np.array([[pathsDist[i][j] for i in sorted(pathsDist[j])] for j in sorted(pathsDist)])

        outQueue.put( (win, dists, copy.deepcopy(pathsPred)) )

def calcBetweenPar(nxGraphs, inQueue, outQueue):
    '''Wrapper to calculate betweeness in parallel.
    
    The betweeness calculations used here only take into account the number of paths 
    passing through a given edge, so no weight are considered.
    
    For every window, the function stores in the output queue an ordered dict of node pairs with betweenness higher than zero.
    
    Args:
        nxGraphs (obj) : NetworkX object.
        inQueue (obj) : Multiprocessing queue object for acquiring jobs.
        outQueue (obj) : Multiprocessing queue object for placing results.
    
    '''
    
    while not inQueue.empty():
        
        win = inQueue.get()
        
        # Calc all betweeness in entire system.
        ### IMPORTANT!
        # For the betweeness, we only care about the number of shortests paths 
        #   passing through a given edge, so no weight are considered.
        btws = nxbetweenness(nxGraphs[win], weight=None)
        
        # Creates an ordered dict of pairs with betweenness higher than zero.
        btws = {k:btws[k] for k in btws.keys() if btws[k] > 0}
        btws = OrderedDict(sorted(btws.items(), key=lambda t: t[1], reverse=True))
        
        outQueue.put((win, btws))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# @author: melomcr


import numpy as np
import cython

# For generalized correlations
from numba import jit

##################################################
##################################################

## Auxiliary functions for calculation of correlation coefficients.

@cython.cfunc
@cython.returns(cython.void)
@cython.locals(numNodes=cython.int, numDims=cython.int, traj="np.ndarray[np.float_t, ndim=3]")
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
def standVarsC(traj, numNodes, numDims):
    '''Standardize variables in trajectory data.
    
    This function prepares the trajectory for the estimation of mutual information coefficients.
    This calculation assumes that the input trajectory of random variables (in this case, positions of atoms) is provided in a variable-dimension-step format (or "atom by x/y/z by frame" for molecular dynamics data). 
        
    .. note:: Please refer to :py:func:`prepMIc` for details about the data conversion process. Please refer to :py:func:`calcMIRnumba2var` for details about the calculation of mutual information coefficients.
    
    Args:
        traj (obj): NumPy array with trajectory information.
        numNodes (int): Number of nodes in the network.
        numDims (int): Number of dimensions in trajectory data (usually 3 dimensions, for X,Y,Z coordinates).
    
    '''
    
    # Standardize variables
    for atm in range(numNodes):
        for dim in range(numDims):
            # Noromalize each dimension.
            traj[atm, dim, :] = (traj[atm, dim, :] - traj[atm, dim, :].mean())/ traj[atm, dim, :].std()

            # Offset all data by minimum normalilzed value.
            traj[atm, dim, :] -= traj[atm, dim, :].min()

@cython.cfunc
@cython.returns(cython.void)
@cython.locals(traj="np.ndarray[np.float_t, ndim=3]",
               beg=cython.int, 
               end=cython.int, 
               numNodes=cython.int, 
               numDims=cython.int)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
def prepMIc(universe, traj, beg, end, numNodes, numDims):
    '''Standardize variables in trajectory data.
    
    This function stores the trajectory data in a new format to accelerate the estimation of mutual information coefficients.
    The estimation of mutual information coefficients assumes that the input trajectory of random variables (in this case, positions of atoms) is provided in a variable-dimension-step format (or "atom by x/y/z by frame" for molecular dynamics data). However, the standard MDAnalysis trajectory format if "frame by atom by x/y/z".
    
    This function also calls a dedicated function to standardize the atom position data, also necessary for mutual information estimation.
        
    .. note:: Please refer to :py:func:`standVarsC` for details about the data standardization process. Please refer to :py:func:`calcMIRnumba2var` for details about the calculation of mutual information coefficients.
    
    Args:
        universe (obj): MDAnalysis universe object containing all trajectory information.
        traj (obj): NumPy array where trajectory information will be stored.
        beg (int): Initial trajectory frame to be used for analysis.
        end (int): Final trajectory frame to be used for analysis.
        numNodes (int): Number of nodes in the network.
        numDims (int): Number of dimensions in trajectory data (usually 3 dimensions, for X,Y,Z coordinates).
    
    '''
    
    # Copy trajectory
    for frameIndx, ts in enumerate(universe.trajectory[beg:end]):
    
        for atmIndx in range(numNodes):
            for dim in range(numDims):
                traj[atmIndx, dim, frameIndx] = ts.positions[atmIndx,dim]
    
    standVarsC(traj, numNodes, numDims)
 
#
# This takes ~4.52 ms ± 43.1 µs for 250 frames
#
@jit('f8(f8[:,:,:], i4, i8, i4, f8[:], f8[:])', nopython=True)
def calcMIRnumba2var(traj, numFrames, numDims, kNeighb, psi, phi):
    '''Calculate mutual information coefficients.
    
    This function estimates the mutual information coefficient based on `Kraskov et. al. (2004) <https://doi.org/10.1103/PhysRevE.69.066138>`_, using the rectangle method. This implementation is hardcoded for 2 variables, to maximize efficiency, and is Just-In-Time (JIT) compiled using `Numba`_.
    
    .. _Numba: https://numba.pydata.org/

    This calculation assumes that the input trajectory of two random variables (in this case, positions of two atoms) is provided in a variable-dimension-step format (or "atom by x/y/z by frame" for molecular dynamics data). It also assumes that all trajectory data has been standardized (see :py:func:`standVarsC`).
    
    Args:
        traj (obj): NumPy array with trajectory information.
        numFrames (int): Number of trajectory frames in the current window.
        numDims (int): Number of dimensions in trajectory data (usually 3 dimensions, for X,Y,Z coordinates).
        kNeighb (int): Parameter used for mutual information estimation.
        psi (float): Pre-calculated parameter used for mutual information estimation.
        phi (float): Pre-calculated parameter used for mutual information estimation.
    
    '''
    
    dxy = 0
    
    diffX = np.zeros(numFrames, dtype=np.float64)
    diffY = np.zeros(numFrames, dtype=np.float64)
    tmpDiff = np.zeros(numFrames, dtype=np.float64)
    sortIndx = np.zeros(numFrames, dtype=np.int64)
    
    for step in range(numFrames):
        diffX.fill(0)
        diffY.fill(0)
        tmpDiff.fill(0)
        sortIndx.fill(0)

        for d in range(numDims):
            
            tmpDiff = np.abs( traj[0,d,:] - traj[0,d,step] )
            
            diffX = np.where( diffX > tmpDiff, diffX, tmpDiff )
            
            tmpDiff = np.abs( traj[1,d,:] - traj[1,d,step] )
            
            diffY = np.where( diffY > tmpDiff, diffY, tmpDiff )
        
        # Create an array with indices of sorted distance arrays
        sortIndx = np.argsort( np.where(diffX > diffY, diffX, diffY))
        
        epsx = 0
        epsy = 0
        
        # Get the maximum diatance in each dimention, for each variable,
        #  among k nearest neighbors.
        # We add one to the count to include the k-th neighbour, as the first index
        #  in the list it the frame itself.
        for kindx in range(1, kNeighb+1):
            
            for d in range(numDims):
                # For variable "i"
                dist = np.abs( traj[0,d,step] - traj[0, d, sortIndx[kindx]] )
                
                if epsx < dist :
                    epsx = dist

                # For variable "j"
                dist = np.abs( traj[1,d,step] - traj[1, d, sortIndx[kindx]] )
                
                if epsy < dist :
                    epsy = dist
        
        # Count the number of frames in which the point is within "eps-" distance from
        #   the position in frame "step". Subtract one so not to count the frame itself.
        nx = len(np.nonzero( diffX <= epsx )[0]) -1
        ny = len(np.nonzero( diffY <= epsy )[0]) -1
        
        dxy += psi[nx] + psi[ny];
    
    dxy /= numFrames
    
    # Mutual Information R
    return psi[numFrames] + phi[kNeighb] - dxy

def calcCorProc(traj, winLen, psi, phi, numDims, kNeighb, inQueue, outQueue):
    '''Process for parallel calculation of generalized correlation coefficients.
    
    This function serves as a wrapper and manager for the calculation of generalized correlation coefficients. It uses Python's `multiprocessing` module to launch `Processes` for parallel execution, where each process uses two `multiprocessing` queues to manage job acquisition and saving results. 
    For each calculation, the job acquisition queue passes a trajectories of a pair of atoms, while the output queue stores the generalized correlation coefficient. 
    The generalized correlation coefficient is claculated using a mutual information coefficient which is estimated using an optimized function.
    
    .. note:: Please refer to :py:func:`calcMIRnumba2var` for details about the calculation of mutual information coefficients.
    
    Args:
        traj (obj): NumPy array with trajectory information.
        winLen (int): Number of trajectory frames in the current window.
        psi (float): Pre-calculated parameter used for mutual information estimation.
        phi (float): Pre-calculated parameter used for mutual information estimation.
        numDims (int): Number of dimensions in trajectory data (usually 3 dimensions, for X,Y,Z coordinates).
        kNeighb (int): Parameter used for mutual information estimation.
        inQueue (obj) : Multiprocessing queue object for acquiring jobs.
        outQueue (obj) : Multiprocessing queue object for placing results.
        
    '''
    
    # While we still hape elements to process, get a new pair of nodes. 
    while not inQueue.empty():
        
        atmList = inQueue.get()
        
        # Calls the Numba-compiled function.
        corr = calcMIRnumba2var(traj[atmList, :, :], winLen, numDims, kNeighb, psi, phi)
        
        # Assures that the Mutual Information estimate is not lower than zero.
        corr = max(0, corr)
        
        # Determine generalized correlation coeff from the Mutual Information
        corr = np.sqrt(1-np.exp(-corr*(2.0/3)));
            
        #return (atmList, corr)
        outQueue.put( (atmList, corr) )







#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# @author: melomcr


import numpy as np

from MDAnalysis.analysis            import distances            as mdadist 
from MDAnalysis.lib                 import distances            as mdalibdist
from numba import jit

import cython

# For timing and benchmarks
from timeit import default_timer as timer
from datetime import timedelta

MODE_ALL    = 0
MODE_CAPPED = 1

@jit('i8(i4, i8, i4)', nopython=True)
def getLinIndexNumba(src, trgt, n):
    '''Conversion from 2D matrix indices to 1D triangular.
    
    Converts from 2D matrix indices to 1D (n*(n-1)/2) unwrapped triangular matrix index.
    This version of the function is JIT compiled using Numba.
    
    Args:
        src (int): Source node.
        trg (int): Target node.
        n (int): Dimension of square matrix
    
    Returns:
        int: 1D index in unwrapped triangular matrix.
    
    '''
    
    # based on https://stackoverflow.com/questions/27086195/linear-index-upper-triangular-matrix
    k = (n*(n-1)/2) - (n-src)*((n-src)-1)/2 + trgt - src - 1.0
    return k

@jit('void(i8, i8, f8[:], i8[:], i8[:], i8[:], f8[:])', nopython=True)
def atmToNodeDist(numNodes, nAtoms, tmpDists, atomToNode, nodeGroupIndicesNP, nodeGroupIndicesNPAux, nodeDists):
    '''Translates MDAnalysis distance calculation to node distance matrix.
    
    This function is JIT compiled by Numba to optimize the search for shortest cartesian distances between atoms in different node groups . It relies on the results of MDAnalysis' `self_distance_array` calculation, stored in a 1D NumPy array of shape (n*(n-1)/2,), which acts as an unwrapped triangular matrix.
    
    The pre-allocated triangular matrix passed as an argument to this function is used to store the shortest cartesian distance between each pair of nodes.
    
    This is intended as an analysis tool to allow the comparison of network distances and cartesian distances. It is similar to :py:func:`calcContactC`, which is optimized for contact detection.
    
    Args:
        numNodes (int): Number of nodes in the system.
        nAtoms (int) : Number of atoms in atom groups represented by system nodes. Usually hydrogen atoms are not included in contact detection, and are not present in atom groups.
        atomToNode (obj) : NumPy array that maps atoms in atom groups to their respective nodes.
        nodeGroupIndicesNP (obj) : NumPy array with atom indices for all atoms in each node group.
        nodeGroupIndicesNPAux (obj) : Auxiliary NumPy array with the indices of the first atom in each atom group, as listed in `nodeGroupIndicesNP`.
        nodeDists (obj) : Pre-allocated array to store cartesian distances.
        backend (str) : Controls how MDAnalysis will perform its distance calculations. Options are  `serial` and `openmp`.
        
    '''
    
    maxVal = np.finfo(np.float64).max
    
    # Initialize with maximum possible value of a float64
    tmpDistsAtms = np.full( nAtoms, maxVal, dtype=np.float64 )
    
    # We iterate untill we have only one node left
    for i in range(numNodes - 1):
        
        # Initializes the array for this node.
        tmpDistsAtms.fill(maxVal)
        
        # index of first atom in node "i"
        nodeAtmIndx = nodeGroupIndicesNPAux[i]
        # index of first atom in node "i+1"
        nodeAtmIndxNext = nodeGroupIndicesNPAux[i+1]
        
        # Gets first atom in next node
        nextIfirst = nodeGroupIndicesNP[nodeAtmIndxNext]
        
        # Per node:  Iterate over atoms from node i
        for nodeI_k in nodeGroupIndicesNP[nodeAtmIndx:nodeAtmIndxNext]:
            
            # Go from 2D indices to 1D (n*(n-1)/2) indices:
            j1 = getLinIndexNumba(nodeI_k, nextIfirst, nAtoms)
            jend = j1 + (nAtoms - nextIfirst)
            
            # Gets the shortest distance between atoms in different nodes
            tmpDistsAtms[nextIfirst:] = np.where(tmpDists[j1: jend] < tmpDistsAtms[nextIfirst:], 
                     tmpDists[j1: jend], 
                     tmpDistsAtms[nextIfirst:])
            
        for pairNode in range(i+1, numNodes):
            
            # Access the shortests distances between atoms from "pairNode" and the current node "i"
            minDist = np.min(tmpDistsAtms[ np.where(atomToNode == pairNode) ])
            
            # Go from 2D node indices to 1D (numNodes*(numNodes-1)/2) indices:
            ijLI = getLinIndexNumba(i, pairNode, numNodes)
            
            nodeDists[ijLI] = minDist
    
# High memory usage (nAtoms*(nAtoms-1)/2), calcs all atom distances at once.
# We use self_distance_array and iterate over the trajectory.
# https://www.mdanalysis.org/mdanalysis/documentation_pages/analysis/distances.html
def calcDistances(selection, numNodes, nAtoms, atomToNode,  cutoffDist,
                       nodeGroupIndicesNP, nodeGroupIndicesNPAux, nodeDists, backend="serial", distMode=MODE_ALL, verbose=0):
    '''Executes MDAnalysis atom distance calculation and node cartesian distance calculation.
    
    This function is a wrapper for two optimized atomic distance calculation and node distance calculation calls.
    The first is one of MDAnalysis' atom distance calculation functions (either `self_distance_array` or `self_capped_distance`). The second is the internal :py:func:`atmToNodeDist`.
    All results are stored in pre-allocated NumPy arrays.
    
    This is intended as an analysis tool to allow the comparison of network distances and cartesian distances. It is similar to :py:func:`getContactsC`, which is optimized for contact detection.
    
    Args:
        selection (str) : Atom selection for the system being analyzed.
        numNodes (int): Number of nodes in the system.
        nAtoms (int) : Number of atoms in atom groups represented by system nodes. Usually hydrogen atoms are not included in contact detection, and are not present in atom groups.
        atomToNode (obj) : NumPy array that maps atoms in atom groups to their respective nodes.
        cutoffDist (float): Distance cutoff used to capp distance calculations. 
        nodeGroupIndicesNP (obj) : NumPy array with atom indices for all atoms in each node group.
        nodeGroupIndicesNPAux (obj) : Auxiliary NumPy array with the indices of the first atom in each atom group, as listed in `nodeGroupIndicesNP`.
        nodeDists (obj) : Pre-allocated array to store cartesian distances.
        backend (str) : Controls how MDAnalysis will perform its distance calculations. Options are  `serial` and `openmp`. This option is ignored if the ditance mode is not "all".
        distMode (str): Distance calculation method. Options are 0 (for mode "all") and 1 (for mode "capped").
        verbose (int): Controls informational output.
        
    '''
    
    if verbose:
        print("There are {} nodes and {} atoms in this system.".format(numNodes, nAtoms))
    
    if distMode == MODE_ALL:
        
        if verbose:
            print("creating array with {} elements...".format(int(nAtoms*(nAtoms-1)/2)))
            start = timer()
            
        tmpDists = np.zeros( int(nAtoms*(nAtoms-1)/2), dtype=np.float64 )
        
        if verbose:
            end = timer()
            print("Time for matrix:", timedelta(seconds=end-start))
        
        if verbose:
            print("running self_distance_array...")
            start = timer()
        
        # serial vs OpenMP
        mdadist.self_distance_array(selection.positions, result=tmpDists, backend=backend)
        
        if verbose:
            end = timer()
            print("Time for contact calculation:", timedelta(seconds=end-start))
        
    if distMode == MODE_CAPPED:
        
        if verbose:
            print("creating array with {} elements...".format(int(nAtoms*(nAtoms-1)/2)))
            start = timer()
        
        tmpDists =  np.full( int(nAtoms*(nAtoms-1)/2), cutoffDist*2, dtype=float )
        
        if verbose:
            end = timer()
            print("Time for matrix:", timedelta(seconds=end-start))
        
        if verbose:
            print("running self_capped_distance...")
            start = timer()
            
        # method options are: 'bruteforce' 'nsgrid' 'pkdtree'
        pairs, distances = mdalibdist.self_capped_distance(selection.positions, max_cutoff=cutoffDist, 
                                    min_cutoff=None, box=None, method='pkdtree', return_distances=True)
        
        if verbose:
            end = timer()
            print("Time for contact calculation:", timedelta(seconds=end-start))
            
            print("Found {} pairs and {} distances".format(len(pairs), len(distances)) )
        
        if verbose:
            print("loading distances in array...")
            start = timer()
            if verbose > 1:
                startLoop = timer()
        
        for k in range(len(pairs)):
            i,j = pairs[k]
            
            if verbose > 1:
                if not k % 1000:
                    print("Loaded {} distances.".format(k))
                    print("Time for {} distances: {}".format(k, timedelta(seconds=timer()-startLoop)))
                    startLoop = timer()
                
            # Go from 2D node indices to 1D (numNodes*(numNodes-1)/2) indices:
            ijLI = getLinIndexNumba(i, j, nAtoms)
            tmpDists[ ijLI ] = distances[k]
            
        if verbose:
            end = timer()
            print("Time for loading distances:", timedelta(seconds=end-start))
            
            print("running atmToNodeDist...")
            start = timer()
    
    # Translate atoms distances in minimum node distance.
    atmToNodeDist(numNodes, nAtoms, tmpDists, atomToNode, nodeGroupIndicesNP, nodeGroupIndicesNPAux, nodeDists)
    
    if verbose:
        end = timer()
        print("Time for atmToNodeDist:", timedelta(seconds=end-start))

@cython.cfunc
@cython.returns(cython.int)
@cython.locals(src=cython.int, trgt=cython.int, n=cython.int)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
def getLinIndexC(src, trgt, n):
    '''Conversion from 2D matrix indices to 1D triangular.
    
    Converts from 2D matrix indices to 1D (n*(n-1)/2) unwrapped triangular matrix index.
    This version of the function is compiled using Cython.
    
    Args:
        src (int): Source node.
        trg (int): Target node.
        n (int): Dimension of square matrix
    
    Returns:
        int: 1D index in unwrapped triangular matrix.
    
    '''
    
    # based on https://stackoverflow.com/questions/27086195/linear-index-upper-triangular-matrix
    k = (n*(n-1)/2) - (n-src)*((n-src)-1)/2 + trgt - src - 1.0
    return int(k)
    
@cython.cfunc
@cython.returns(cython.void)
@cython.locals(numNodes=cython.int, nAtoms=cython.int, cutoffDist=cython.float,
               tmpDists="np.ndarray[np.float_t, ndim=1]",
               tmpDistsAtms="np.ndarray[np.float_t, ndim=1]",
               contactMat="np.ndarray[np.int_t, ndim=2]",
               atomToNode="np.ndarray[np.int_t, ndim=1]",
               nodeGroupIndicesNP="np.ndarray[np.int_t, ndim=1]",
               nodeGroupIndicesNPAux="np.ndarray[np.int_t, ndim=1]")
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
def calcContactC(numNodes, nAtoms, cutoffDist, 
                 tmpDists, 
                 tmpDistsAtms, 
                 contactMat,
                 atomToNode,
                 nodeGroupIndicesNP,
                 nodeGroupIndicesNPAux):
    '''Translates MDAnalysis distance calculation to node contact matrix.
    
    This function is Cython compiled to optimize the search for nodes in contact. It relies on the results of MDAnalysis' `self_distance_array` calculation, stored in a 1D NumPy array of shape (n*(n-1)/2,), which acts as an unwrapped triangular matrix.
    
    In this function, the distances between all atoms in an atom groups of all pairs of nodes are verified to check if any pair of atoms were closer than a cutoff distance. This is done for all pairs of nodes in the system, and all frames in the trajectory. The pre-allocated contact matrix passed as an argument to this function is used to store the number of frames where each pair of nodes had at least one contact.
    
    Args:
        numNodes (int): Number of nodes in the system.
        nAtoms (int) : Number of atoms in atom groups represented by system nodes. Usually hydrogen atoms are not included in contact detection, and are not present in atom groups.
        cutoffDist (float) : Distance at which atoms are no longer considered 'in contact'.
        tmpDists (obj) : Temporary pre-allocated NumPy array with atom distances. This is the result of MDAnalysis `self_distance_array` calculation.
        tmpDistsAtms (obj) : Temporary pre-allocated NumPy array to store the shortest distance between atoms in different nodes.
        contactMat (obj) : Pre-allocated NumPy matrix where node contacts will be stored.
        atomToNode (obj) : NumPy array that maps atoms in atom groups to their respective nodes.
        nodeGroupIndicesNP (obj) : NumPy array with atom indices for all atoms in each node group.
        nodeGroupIndicesNPAux (obj) : Auxiliary NumPy array with the indices of the first atom in each atom group, as listed in `nodeGroupIndicesNP`.
    
    '''
    
    #cdef int nextIfirst, j1, jend, i, nodeI_k, nodeAtmIndx, nodeAtmIndxNext
    nextIfirst = cython.declare(cython.int)
    
    # Cython types are evaluated as for cdef declarations
    j1: cython.int
    jend: cython.int
    i: cython.int
    nodeI_k: cython.int
    nodeAtmIndx: cython.int
    nodeAtmIndxNext: cython.int
    
    
    # We iterate untill we have only one node left
    for i in range(numNodes - 1):
        
        # Initializes the array for this node.
        tmpDistsAtms.fill(cutoffDist*2)
        
        # index of first atom in node "i"
        nodeAtmIndx = nodeGroupIndicesNPAux[i]
        # index of first atom in node "i+1"
        nodeAtmIndxNext = nodeGroupIndicesNPAux[i+1]
        
        # Gets first atom in next node
        nextIfirst = nodeGroupIndicesNP[nodeAtmIndxNext]
        
        # Per node:  Iterate over atoms from node i
        for nodeI_k in nodeGroupIndicesNP[nodeAtmIndx:nodeAtmIndxNext]:
            
            # Go from 2D indices to 1D (n*(n-1)/2) indices:
            j1 = getLinIndexC(nodeI_k, nextIfirst, nAtoms)
            jend = j1 + (nAtoms - nextIfirst)
            
            # Gets the shortest distance between atoms in different nodes
            tmpDistsAtms[nextIfirst:] = np.where(tmpDists[j1: jend] < tmpDistsAtms[nextIfirst:], 
                     tmpDists[j1: jend], 
                     tmpDistsAtms[nextIfirst:])
        
        # Adds one to the contact to indicate that this frame had a contact.
        contactMat[i, np.unique( atomToNode[ np.where(tmpDistsAtms < cutoffDist)[0] ] )] += 1

# High memory usage (nAtoms*(nAtoms-1)/2), calcs all atom distances at once.
@cython.cfunc
@cython.returns(cython.void)
@cython.locals(numNodes=cython.int, nAtoms=cython.int, cutoffDist=cython.float,
               tmpDists="np.ndarray[np.float_t, ndim=1]",
               tmpDistsAtms="np.ndarray[np.float_t, ndim=1]",
               contactMat="np.ndarray[np.int_t, ndim=2]",
               atomToNode="np.ndarray[np.int_t, ndim=1]",
               nodeGroupIndicesNP="np.ndarray[np.int_t, ndim=1]",
               nodeGroupIndicesNPAux="np.ndarray[np.int_t, ndim=1]")
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
def getContactsC(selection, numNodes, 
                        nAtoms, 
                        cutoffDist, 
                        tmpDists, 
                        tmpDistsAtms, 
                        contactMat,
                        atomToNode,
                        nodeGroupIndicesNP,
                        nodeGroupIndicesNPAux,
                        distMode=MODE_ALL):
    '''Executes MDAnalysis atom distance calculation and node contact detection.
    
    This function is Cython compiled as a wrapper for two optimized distance calculation and contact determination calls.
    The first is MDAnalysis' `self_distance_array`. The second is the internal :py:func:`calcContactC`.
    All results are stored in pre-allocated NumPy arrays.
    
    
    Args:
        selection (str) : Atom selection for the system being analyzed.
        numNodes (int): Number of nodes in the system.
        nAtoms (int) : Number of atoms in atom groups represented by system nodes. Usually hydrogen atoms are not included in contact detection, and are not present in atom groups.
        cutoffDist (float) : Distance at which atoms are no longer considered 'in contact'.
        tmpDists (obj) : Temporary pre-allocated NumPy array with atom distances. This is the result of MDAnalysis `self_distance_array` calculation.
        tmpDistsAtms (obj) : Temporary pre-allocated NumPy array to store the shortest distance between atoms in different nodes.
        contactMat (obj) : Pre-allocated NumPy matrix where node contacts will be stored.
        atomToNode (obj) : NumPy array that maps atoms in atom groups to their respective nodes.
        nodeGroupIndicesNP (obj) : NumPy array with atom indices for all atoms in each node group.
        nodeGroupIndicesNPAux (obj) : Auxiliary NumPy array with the indices of the first atom in each atom group, as listed in `nodeGroupIndicesNP`.
    
    '''
    
    if distMode == MODE_ALL:
        # serial vs OpenMP
        mdadist.self_distance_array(selection.positions, result=tmpDists, backend='openmp')
    
    if distMode == MODE_CAPPED:
        # method options are: 'bruteforce' 'nsgrid' 'pkdtree'
        pairs, distances = mdalibdist.self_capped_distance(selection.positions, max_cutoff=cutoffDist, min_cutoff=None, box=None,
                                    method='pkdtree', return_distances=True)
        
        for k, [i, j] in enumerate(pairs):
            # Go from 2D node indices to 1D (nAtoms*(nAtoms-1)/2) indices:
            ijLI = getLinIndexC(i, j, nAtoms)
            tmpDists[ ijLI ] = distances[k]
    
    calcContactC(numNodes, nAtoms, cutoffDist, tmpDists, tmpDistsAtms, 
                 contactMat, atomToNode, nodeGroupIndicesNP, nodeGroupIndicesNPAux)


